@@ -40,11 +40,18 @@
 #include <panel-applet-gconf.h>
 
 #include "gdm-user-manager.h"
-#include "gdm-user-menu-item.h"
+#include "gdm-entry-menu-item.h"
 
 #define LOCKDOWN_DIR    "/desktop/gnome/lockdown"
 #define LOCKDOWN_KEY    LOCKDOWN_DIR "/disable_user_switching"
 
+enum {
+        GDM_USER_STATUS_AVAILABLE,
+        GDM_USER_STATUS_BUSY,
+        GDM_USER_STATUS_INVISIBLE,
+        GDM_USER_STATUS_AWAY,
+};
+
 typedef struct _GdmAppletData
 {
         PanelApplet    *applet;
@@ -56,17 +63,20 @@ typedef struct _GdmAppletData
         GtkWidget      *menubar;
         GtkWidget      *menuitem;
         GtkWidget      *menu;
-        GtkWidget      *separator_item;
+        GtkWidget      *user_item;
+        GtkWidget      *control_panel_item;
+        GtkWidget      *account_item;
+        GtkWidget      *lock_screen_item;
         GtkWidget      *login_screen_item;
-        GSList         *items;
+        GtkWidget      *quit_session_item;
 
-        gboolean        active_only;
+        gboolean        has_other_users;
 
         guint           client_notify_lockdown_id;
 
+        guint           current_status;
         guint           user_icon_changed_id;
         guint           user_notify_id;
-        GQuark          user_menu_item_quark;
         gint8           pixel_size;
         gint            panel_size;
         GtkIconSize     icon_size;
@@ -78,9 +88,8 @@ typedef struct _SelectorResponseData
         GtkRadioButton *radio;
 } SelectorResponseData;
 
-static GtkTooltips *tooltips = NULL;
-
-static void reset_icon (GdmAppletData *adata);
+static void reset_icon   (GdmAppletData *adata);
+static void update_label (GdmAppletData *adata);
 
 static gboolean applet_factory (PanelApplet   *applet,
                                 const char    *iid,
@@ -95,7 +104,7 @@ PANEL_APPLET_BONOBO_FACTORY ("OAFIID:GNOME_FastUserSwitchApplet_Factory",
 static void
 about_me_cb (BonoboUIComponent *ui_container,
              gpointer           data,
-             const char       *cname)
+             const char        *cname)
 {
         GError *err;
 
@@ -135,7 +144,7 @@ menubar_button_press_event_cb (GtkWidget      *menubar,
 static void
 about_cb (BonoboUIComponent *ui_container,
           gpointer           data,
-          const char       *cname)
+          const char        *cname)
 {
         static const char *authors[] = {
                 "James M. Cape <jcape@ignore-your.tv>",
@@ -181,7 +190,7 @@ about_cb (BonoboUIComponent *ui_container,
 static void
 admin_cb (BonoboUIComponent *ui_container,
           gpointer           data,
-          const char       *cname)
+          const char        *cname)
 {
 #ifdef USERS_ADMIN
         char   **args;
@@ -256,9 +265,11 @@ applet_style_set_cb (GtkWidget *widget,
                      gpointer   data)
 {
         BonoboUIComponent *component;
-        GdkScreen *screen;
-        GtkIconTheme *theme;
-        gint width, height, icon_size;
+        GdkScreen         *screen;
+        GtkIconTheme      *theme;
+        int                width;
+        int                height;
+        int                icon_size;
 
         if (gtk_widget_has_screen (widget)) {
                 screen = gtk_widget_get_screen (widget);
@@ -293,13 +304,10 @@ applet_change_background_cb (PanelApplet               *applet,
                              PanelAppletBackgroundType  type,
                              GdkColor                  *color,
                              GdkPixmap                 *pixmap,
-                             gpointer                   data)
+                             GdmAppletData             *adata)
 {
-        GdmAppletData *adata;
-        GtkRcStyle    *rc_style;
-        GtkStyle      *style;
-
-        adata = data;
+        GtkRcStyle *rc_style;
+        GtkStyle   *style;
 
         gtk_widget_set_style (adata->menubar, NULL);
         rc_style = gtk_rc_style_new ();
@@ -333,11 +341,10 @@ applet_change_background_cb (PanelApplet               *applet,
  * Copyright (C) 2000 Helix Code, Inc.
  */
 static gboolean
-applet_key_press_event_cb (GtkWidget   *widget,
-                           GdkEventKey *event,
-                           gpointer     data)
+applet_key_press_event_cb (GtkWidget     *widget,
+                           GdkEventKey   *event,
+                           GdmAppletData *adata)
 {
-        GdmAppletData *adata;
         GtkMenuShell *menu_shell;
 
         switch (event->keyval) {
@@ -347,7 +354,6 @@ applet_key_press_event_cb (GtkWidget   *widget,
         case GDK_Return:
         case GDK_space:
         case GDK_KP_Space:
-                adata = data;
                 menu_shell = GTK_MENU_SHELL (adata->menubar);
                 /*
                  * We need to call _gtk_menu_shell_activate() here as is done in
@@ -475,7 +481,6 @@ gdm_applet_data_free (GdmAppletData *adata)
         }
         g_object_unref (adata->client);
         g_object_unref (adata->manager);
-        g_object_unref (tooltips);
 
         g_free (adata);
 }
@@ -491,12 +496,8 @@ gdm_applet_data_free (GdmAppletData *adata)
 static gboolean
 menubar_expose_event_cb (GtkWidget      *widget,
                          GdkEventExpose *event,
-                         gpointer        data)
+                         GdmAppletData  *adata)
 {
-        GdmAppletData *adata;
-
-        adata = data;
-
         if (GTK_WIDGET_HAS_FOCUS (adata->applet))
                 gtk_paint_focus (widget->style, widget->window, GTK_WIDGET_STATE (widget),
                                  NULL, widget, "menu-applet", 0, 0, -1, -1);
@@ -504,128 +505,15 @@ menubar_expose_event_cb (GtkWidget      *widget,
         return FALSE;
 }
 
-static gint
-sort_menu_comparedatafunc (gconstpointer a,
-                           gconstpointer b,
-                           gpointer      data)
-{
-        GdmAppletData *adata;
-        gboolean a_is_user, b_is_user;
-
-        a_is_user = GDM_IS_USER_MENU_ITEM (a);
-        b_is_user = GDM_IS_USER_MENU_ITEM (b);
-
-        if (a_is_user && !b_is_user) {
-                return -1;
-        }
-
-        if (b_is_user && !a_is_user) {
-                return 1;
-        }
-
-        if (a_is_user && b_is_user) {
-                return gdm_user_collate (gdm_user_menu_item_get_user ((GdmUserMenuItem *) a),
-                                         gdm_user_menu_item_get_user ((GdmUserMenuItem *) b));
-        }
-
-        adata = data;
-        if (a == adata->separator_item) {
-                return -1;
-        }
-
-        if (b == adata->separator_item) {
-                return 1;
-        }
-
-        if (a == adata->login_screen_item) {
-                return -1;
-        }
-
-        if (b == adata->login_screen_item) {
-                return 1;
-        }
-
-        return 0;
-}
-
 static void
-sort_menu (GdmAppletData *adata)
+menu_style_set_cb (GtkWidget     *menu,
+                   GtkStyle      *old_style,
+                   GdmAppletData *adata)
 {
-        GSList *items;
-        guint   n_items;
-        guint   n_rows;
-        guint   n_cols;
-        guint   row;
-        guint   column;
-        guint   count;
-        int     screen_height;
-
-        if (!gtk_widget_has_screen (adata->menu)) {
-                return;
-        }
-
-        adata->items = g_slist_sort_with_data (adata->items,
-                                               sort_menu_comparedatafunc,
-                                               adata);
-
-        screen_height = gdk_screen_get_height (gtk_widget_get_screen (adata->menu));
-
-        n_items = 0;
-        items = adata->items;
-        while (items) {
-                if (GTK_WIDGET_VISIBLE (items->data)) {
-                        n_items++;
-                }
-
-                items = items->next;
-        }
-
-        /* FIXME: Do a better job of figuring out exactly how big the menuitems are */
-        n_rows = (gdouble) screen_height / (gdouble) (adata->pixel_size + 16);
-        n_cols = (gdouble) n_items / (gdouble) n_rows;
-        n_rows = (gdouble) n_items / (gdouble) (n_cols + 1);
-
-        row = 0;
-        column = 0;
-        count = 0;
-        items = adata->items;
-        while (items != NULL) {
-                if (GTK_WIDGET_VISIBLE (items->data)) {
-                        gtk_menu_attach (GTK_MENU (adata->menu),
-                                         items->data,
-                                         column,
-                                         column + 1,
-                                         row,
-                                         row + 1);
-                        row++;
-                        if (row > n_rows) {
-                                row = 0;
-                                column++;
-                        }
-
-                        /* Just re-attaching them doesn't alter the order you get them
-                         * in when you move through with the arrow keys, though; we
-                         * have to set that explicitly.
-                         */
-                        gtk_menu_reorder_child (GTK_MENU (adata->menu),
-                                                items->data,
-                                                count++);
-                }
-
-                items = items->next;
-        }
-}
-
-static void
-menu_style_set_cb (GtkWidget *menu,
-                   GtkStyle  *old_style,
-                   gpointer   data)
-{
-        GdmAppletData *adata;
         GtkSettings *settings;
-        gint width, height;
+        int          width;
+        int          height;
 
-        adata = data;
         adata->icon_size = gtk_icon_size_from_name ("panel-menu");
 
         if (adata->icon_size == GTK_ICON_SIZE_INVALID) {
@@ -644,53 +532,29 @@ menu_style_set_cb (GtkWidget *menu,
         } else {
                 adata->pixel_size = MAX (width, height);
         }
-
-        sort_menu (adata);
 }
 
 static void
-menuitem_destroy_cb (GtkWidget *menuitem,
-                     gpointer   data)
+menuitem_style_set_cb (GtkWidget     *menuitem,
+                       GtkStyle      *old_style,
+                       GdmAppletData *adata)
 {
-        GdmAppletData *adata;
-        GSList        *li;
+        GtkWidget *image;
 
-        adata = data;
-
-        if (GDM_IS_USER_MENU_ITEM (menuitem)) {
-                GdmUser *user;
-
-                user = gdm_user_menu_item_get_user (GDM_USER_MENU_ITEM (menuitem));
-                if (user != NULL) {
-                        g_object_set_qdata (G_OBJECT (user),
-                                            adata->user_menu_item_quark, NULL);
-                }
-        }
-
-        g_debug ("Menuitem destroyed - removing");
-        li = g_slist_find (adata->items, menuitem);
-        adata->items = g_slist_delete_link (adata->items, li);
-        sort_menu (adata);
-}
-
-static void
-menuitem_style_set_cb (GtkWidget *menuitem,
-                       GtkStyle  *old_style,
-                       gpointer   data)
-{
-        GdmAppletData *adata;
-
-        adata = data;
-
-        if (GDM_IS_USER_MENU_ITEM (menuitem)) {
-                gdm_user_menu_item_set_icon_size (GDM_USER_MENU_ITEM (menuitem),
-                                                   adata->pixel_size);
+        if (GDM_IS_ENTRY_MENU_ITEM (menuitem)) {
         } else {
-                GtkWidget *image;
                 const char *icon_name;
 
                 if (menuitem == adata->login_screen_item) {
                         icon_name = "gdm";
+                } else if (menuitem == adata->lock_screen_item) {
+                        icon_name = "system-lock-screen";
+                } else if (menuitem == adata->quit_session_item) {
+                        icon_name = "system-log-out";
+                } else if (menuitem == adata->account_item) {
+                        icon_name = "user-info";
+                } else if (menuitem == adata->control_panel_item) {
+                        icon_name = "preferences-desktop";
                 } else {
                         icon_name = GTK_STOCK_MISSING_IMAGE;
                 }
@@ -702,19 +566,14 @@ menuitem_style_set_cb (GtkWidget *menuitem,
         }
 }
 
-
 static void
 user_notify_display_name_cb (GObject       *object,
                              GParamSpec    *pspec,
                              GdmAppletData *adata)
 {
-        GtkWidget *label;
-
-        label = gtk_bin_get_child (GTK_BIN (adata->menuitem));
-        gtk_label_set_text (GTK_LABEL (label), gdm_user_get_real_name (GDM_USER (object)));
+        update_label (adata);
 }
 
-
 /* Called every time the menu is displayed (and also for some reason
  * immediately it's created, which does no harm). All we have to do
  * here is kick off a request to GDM to let us know which users are
@@ -728,13 +587,6 @@ menu_expose_cb (GtkWidget *menu,
         return FALSE;
 }
 
-static void
-switch_to_user_session (GdmAppletData *adata,
-                        GdmUser       *user)
-{
-        gdm_user_manager_activate_user_session (adata->manager, user);
-}
-
 static void
 maybe_lock_screen (GdmAppletData *adata)
 {
@@ -825,7 +677,7 @@ do_switch (GdmAppletData *adata,
 
         num_sessions = gdm_user_get_num_sessions (user);
         if (num_sessions > 0) {
-                switch_to_user_session (adata, user);
+                gdm_user_manager_activate_user_session (adata->manager, user);
         } else {
                 gdm_user_manager_goto_login_session (adata->manager);
         }
@@ -834,86 +686,23 @@ do_switch (GdmAppletData *adata,
 }
 
 static void
-user_item_activate_cb (GtkWidget     *menuitem,
-                       GdmAppletData *adata)
+update_switch_user (GdmAppletData *adata)
 {
-        GdmUserMenuItem *item;
-        GdmUser         *user;
+        GSList *users;
 
-        item = GDM_USER_MENU_ITEM (menuitem);
-        user = gdm_user_menu_item_get_user (item);
-
-        do_switch (adata, user);
-}
-
-static void
-update_user_item_visibility (GdmAppletData  *adata,
-                             GdmUser        *user)
-{
-        GtkWidget *menuitem;
-
-        g_debug ("Updating menu item visibility for %s", gdm_user_get_user_name (user));
-
-        menuitem = g_object_get_qdata (G_OBJECT (user), adata->user_menu_item_quark);
-        if (menuitem == NULL) {
-                return;
+        users = gdm_user_manager_list_users (adata->manager);
+        adata->has_other_users = FALSE;
+        if (users != NULL) {
+                adata->has_other_users = (g_slist_length (users) > 1);
         }
+        g_slist_free (users);
 
-        if (adata->active_only) {
-                guint num_sessions;
-
-                num_sessions = gdm_user_get_num_sessions (user);
-                g_debug ("Sessions changed for %s num=%u", gdm_user_get_user_name (user), num_sessions);
-
-                if (num_sessions > 0) {
-                        gtk_widget_show (menuitem);
-                } else {
-                        gtk_widget_hide (menuitem);
-                }
+        if (adata->has_other_users) {
+                gtk_widget_show (adata->login_screen_item);
         } else {
-                gtk_widget_show (menuitem);
+
+                gtk_widget_hide (adata->login_screen_item);
         }
-
-        sort_menu (adata);
-}
-
-static void
-on_user_sessions_changed (GdmUser       *user,
-                          GdmAppletData *adata)
-{
-        g_debug ("Sessions changed for %s", gdm_user_get_user_name (user));
-        update_user_item_visibility (adata, user);
-}
-
-static void
-add_user (GdmAppletData  *adata,
-          GdmUser        *user)
-{
-        GtkWidget *menuitem;
-
-        menuitem = gdm_user_menu_item_new (user);
-        g_object_set_qdata (G_OBJECT (user), adata->user_menu_item_quark, menuitem);
-        g_signal_connect (menuitem,
-                          "style-set",
-                          G_CALLBACK (menuitem_style_set_cb),
-                          adata);
-        g_signal_connect (menuitem,
-                          "destroy",
-                          G_CALLBACK (menuitem_destroy_cb),
-                          adata);
-        g_signal_connect (menuitem,
-                          "activate",
-                          G_CALLBACK (user_item_activate_cb),
-                          adata);
-        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu), menuitem);
-        adata->items = g_slist_prepend (adata->items, menuitem);
-
-        gtk_widget_show (menuitem);
-
-        g_signal_connect (user,
-                          "sessions-changed",
-                          G_CALLBACK (on_user_sessions_changed),
-                          adata);
 }
 
 static void
@@ -921,39 +710,249 @@ on_manager_user_added (GdmUserManager *manager,
                        GdmUser        *user,
                        GdmAppletData  *adata)
 {
-        add_user (adata, user);
-        update_user_item_visibility (adata, user);
-        sort_menu (adata);
+        update_switch_user (adata);
 }
 
 static void
-on_manager_user_is_logged_in_changed (GdmUserManager *manager,
-                                      GdmUser        *user,
-                                      GdmAppletData  *adata)
+on_manager_user_removed (GdmUserManager *manager,
+                         GdmUser        *user,
+                         GdmAppletData  *adata)
 {
-        update_user_item_visibility (adata, user);
-        sort_menu (adata);
+        update_switch_user (adata);
 }
 
 static void
-login_screen_activate_cb (GtkMenuItem *item,
-                          gpointer     data)
+on_manager_users_loaded (GdmUserManager *manager,
+                         GdmAppletData  *adata)
 {
-        GdmAppletData *adata;
-        GdmUser       *user;
+        update_switch_user (adata);
+}
+
+static void
+on_user_item_activate (GtkMenuItem   *item,
+                       GdmAppletData *adata)
+{
+        g_signal_stop_emission_by_name (item, "activate");
+}
+
+static void
+on_control_panel_activate (GtkMenuItem   *item,
+                           GdmAppletData *adata)
+{
+        char      *args[2];
+        GError    *error;
+        GdkScreen *screen;
+        gboolean   res;
+
+        args[0] = g_find_program_in_path ("gnome-control-center");
+        if (args[0] == NULL) {
+                return;
+        }
+        args[1] = NULL;
+
+        if (gtk_widget_has_screen (GTK_WIDGET (adata->applet))) {
+                screen = gtk_widget_get_screen (GTK_WIDGET (adata->applet));
+        } else {
+                screen = gdk_screen_get_default ();
+        }
+
+        error = NULL;
+        res = gdk_spawn_on_screen (screen,
+                                   g_get_home_dir (),
+                                   args,
+                                   NULL,
+                                   0,
+                                   NULL,
+                                   NULL,
+                                   NULL,
+                                   &error);
+        if (! res) {
+                g_warning (_("Can't lock screen: %s"), error->message);
+                g_error_free (error);
+        }
+
+        g_free (args[0]);
+}
+
+static void
+on_account_activate (GtkMenuItem   *item,
+                     GdmAppletData *adata)
+{
+        char      *args[2];
+        GError    *error;
+        GdkScreen *screen;
+        gboolean   res;
+
+        args[0] = g_find_program_in_path ("gnome-about-me");
+        if (args[0] == NULL) {
+                return;
+        }
+        args[1] = NULL;
+
+        if (gtk_widget_has_screen (GTK_WIDGET (adata->applet))) {
+                screen = gtk_widget_get_screen (GTK_WIDGET (adata->applet));
+        } else {
+                screen = gdk_screen_get_default ();
+        }
+
+        error = NULL;
+        res = gdk_spawn_on_screen (screen,
+                                   g_get_home_dir (),
+                                   args,
+                                   NULL,
+                                   0,
+                                   NULL,
+                                   NULL,
+                                   NULL,
+                                   &error);
+        if (! res) {
+                g_warning (_("Can't lock screen: %s"), error->message);
+                g_error_free (error);
+        }
+
+        g_free (args[0]);
+}
+
+static void
+on_lock_screen_activate (GtkMenuItem   *item,
+                         GdmAppletData *adata)
+{
+        maybe_lock_screen (adata);
+}
+
+static void
+on_login_screen_activate (GtkMenuItem   *item,
+                          GdmAppletData *adata)
+{
+        GdmUser *user;
 
-        adata = data;
         user = NULL;
 
         do_switch (adata, user);
 }
 
+static void
+on_quit_session_activate (GtkMenuItem   *item,
+                          GdmAppletData *adata)
+{
+        char      *args[3];
+        GError    *error;
+        GdkScreen *screen;
+        gboolean   res;
+
+        args[0] = g_find_program_in_path ("gnome-session-save");
+        if (args[0] == NULL) {
+                return;
+        }
+
+        args[1] = "--logout-dialog";
+        args[2] = NULL;
+
+        if (gtk_widget_has_screen (GTK_WIDGET (adata->applet))) {
+                screen = gtk_widget_get_screen (GTK_WIDGET (adata->applet));
+        } else {
+                screen = gdk_screen_get_default ();
+        }
+
+        error = NULL;
+        res = gdk_spawn_on_screen (screen,
+                                   g_get_home_dir (),
+                                   args,
+                                   NULL,
+                                   0,
+                                   NULL,
+                                   NULL,
+                                   NULL,
+                                   &error);
+        if (! res) {
+                g_warning (_("Can't logout: %s"), error->message);
+                g_error_free (error);
+        }
+
+        g_free (args[0]);
+}
+
+static gboolean
+on_menu_key_press_event (GtkWidget     *widget,
+                         GdkEventKey   *event,
+                         GdmAppletData *adata)
+{
+        GtkWidget *entry;
+
+        entry = gdm_entry_menu_item_get_entry (GDM_ENTRY_MENU_ITEM (adata->user_item));
+
+        gtk_widget_event (entry, (GdkEvent *)event);
+
+        return FALSE;
+}
+
+static void
+set_status (GdmAppletData *adata,
+            guint          status)
+{
+        adata->current_status = status;
+        update_label (adata);
+}
+
+static void
+on_status_available_activate (GtkWidget     *widget,
+                              GdmAppletData *adata)
+{
+        set_status (adata, GDM_USER_STATUS_AVAILABLE);
+}
+
+static void
+on_status_busy_activate (GtkWidget     *widget,
+                         GdmAppletData *adata)
+{
+        set_status (adata, GDM_USER_STATUS_BUSY);
+}
+
+static void
+on_status_invisible_activate (GtkWidget     *widget,
+                              GdmAppletData *adata)
+{
+        set_status (adata, GDM_USER_STATUS_INVISIBLE);
+}
+
+static struct {
+        char    *icon_name;
+        char    *display_name;
+        void    *menu_callback;
+} statuses[] = {
+        { "user-online", N_("Available"), on_status_available_activate },
+        { "user-busy", N_("Busy"), on_status_busy_activate },
+        { "user-invisible", N_("Invisible"), on_status_invisible_activate },
+        { "user-away", N_("Away"), NULL },
+};
+
+static void
+update_label (GdmAppletData *adata)
+{
+        GtkWidget *label;
+        char      *markup;
+
+        label = gtk_bin_get_child (GTK_BIN (adata->menuitem));
+
+        markup = g_strdup_printf ("<b>%s</b> <small>(%s)</small>",
+                                  gdm_user_get_real_name (GDM_USER (adata->user)),
+                                  statuses[adata->current_status].display_name);
+        gtk_label_set_markup (GTK_LABEL (label), markup);
+        g_free (markup);
+}
+
 static void
 create_sub_menu (GdmAppletData *adata)
 {
-        GSList *users;
+        GtkWidget *item;
+        int        i;
+        GSList    *radio_group;
 
         adata->menu = gtk_menu_new ();
+        g_signal_connect (adata->menu,
+                          "key-press-event",
+                          G_CALLBACK (on_menu_key_press_event),
+                          adata);
         gtk_menu_item_set_submenu (GTK_MENU_ITEM (adata->menuitem), adata->menu);
         g_signal_connect (adata->menu, "style-set",
                           G_CALLBACK (menu_style_set_cb), adata);
@@ -961,56 +960,126 @@ create_sub_menu (GdmAppletData *adata)
                           G_CALLBACK (menu_expose_cb), adata);
         gtk_widget_show (adata->menu);
 
-
-        /* This next part populates the list with all the users we currently know
-         * about. For almost all cases, this is the empty list, because we're
-         * asynchronous, and the data hasn't come back from the callback saying who
-         * the users are yet. However, if someone has two GDMs on their toolbars (why,
-         * I have no freaking idea, but bear with me here), the menu of the second
-         * one to be initialised needs to be filled in from the start rather than
-         * depending on getting data from the callback like the first one.
-         */
-        users = gdm_user_manager_list_users (adata->manager);
-        while (users != NULL) {
-                add_user (adata, users->data);
-                update_user_item_visibility (adata, users->data);
-
-                users = g_slist_delete_link (users, users);
-        }
-
+        g_signal_connect (adata->manager,
+                          "users-loaded",
+                          G_CALLBACK (on_manager_users_loaded),
+                          adata);
         g_signal_connect (adata->manager,
                           "user-added",
                           G_CALLBACK (on_manager_user_added),
                           adata);
         g_signal_connect (adata->manager,
-                          "user-is-logged-in-changed",
-                          G_CALLBACK (on_manager_user_is_logged_in_changed),
+                          "user-removed",
+                          G_CALLBACK (on_manager_user_added),
                           adata);
 
-        adata->separator_item = gtk_separator_menu_item_new ();
-        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu), adata->separator_item);
-        g_signal_connect (adata->separator_item, "destroy",
-                          G_CALLBACK (menuitem_destroy_cb), adata);
-        adata->items = g_slist_prepend (adata->items, adata->separator_item);
-        gtk_widget_show (adata->separator_item);
+        adata->user_item = gdm_entry_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
+                               adata->user_item);
+        gtk_widget_show (adata->user_item);
+        g_signal_connect (adata->user_item, "activate",
+                          G_CALLBACK (on_user_item_activate), adata);
 
-        adata->login_screen_item = gtk_image_menu_item_new_with_label (_("Other..."));
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu), item);
+        gtk_widget_show (item);
+
+        radio_group = NULL;
+        for (i = 0; i < G_N_ELEMENTS (statuses); i++) {
+                GtkWidget *hbox;
+                GtkWidget *label;
+                GtkWidget *image;
+                GtkWidget *item;
+
+                if (statuses[i].menu_callback == NULL) {
+                        continue;
+                }
+
+                item = gtk_radio_menu_item_new (radio_group);
+                radio_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (item));
+                hbox = gtk_hbox_new (FALSE, 3);
+                label = gtk_label_new (statuses[i].display_name);
+                gtk_label_set_justify (GTK_LABEL(label), GTK_JUSTIFY_LEFT);
+                gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
+                gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
+                gtk_widget_show (label);
+                image = gtk_image_new_from_icon_name (statuses[i].icon_name, GTK_ICON_SIZE_MENU);
+                gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, TRUE, 0);
+                gtk_widget_show (image);
+                gtk_widget_show (hbox);
+                gtk_container_add (GTK_CONTAINER (item), hbox);
+
+                gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
+                                       item);
+                g_signal_connect (item, "activate",
+                                  G_CALLBACK (statuses[i].menu_callback), adata);
+                gtk_widget_show (item);
+        }
+
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu), item);
+        gtk_widget_show (item);
+
+        adata->account_item = gtk_image_menu_item_new_with_label (_("Account Information..."));
+        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (adata->account_item),
+                                       gtk_image_new ());
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
+                               adata->account_item);
+        g_signal_connect (adata->account_item, "style-set",
+                          G_CALLBACK (menuitem_style_set_cb), adata);
+        g_signal_connect (adata->account_item, "activate",
+                          G_CALLBACK (on_account_activate), adata);
+        gtk_widget_show (adata->account_item);
+
+
+        adata->control_panel_item = gtk_image_menu_item_new_with_label (_("System Preferences..."));
+        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (adata->control_panel_item),
+                                       gtk_image_new ());
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
+                               adata->control_panel_item);
+        g_signal_connect (adata->control_panel_item, "style-set",
+                          G_CALLBACK (menuitem_style_set_cb), adata);
+        g_signal_connect (adata->control_panel_item, "activate",
+                          G_CALLBACK (on_control_panel_activate), adata);
+        gtk_widget_show (adata->control_panel_item);
+
+
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu), item);
+        gtk_widget_show (item);
+
+        adata->lock_screen_item = gtk_image_menu_item_new_with_label (_("Lock Screen"));
+        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (adata->lock_screen_item),
+                                       gtk_image_new ());
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
+                               adata->lock_screen_item);
+        g_signal_connect (adata->lock_screen_item, "style-set",
+                          G_CALLBACK (menuitem_style_set_cb), adata);
+        g_signal_connect (adata->lock_screen_item, "activate",
+                          G_CALLBACK (on_lock_screen_activate), adata);
+        gtk_widget_show (adata->lock_screen_item);
+
+        adata->login_screen_item = gtk_image_menu_item_new_with_label (_("Switch User"));
         gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (adata->login_screen_item),
                                        gtk_image_new ());
         gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
                                adata->login_screen_item);
         g_signal_connect (adata->login_screen_item, "style-set",
                           G_CALLBACK (menuitem_style_set_cb), adata);
-        g_signal_connect (adata->login_screen_item, "destroy",
-                          G_CALLBACK (menuitem_destroy_cb), adata);
         g_signal_connect (adata->login_screen_item, "activate",
-                          G_CALLBACK (login_screen_activate_cb), adata);
-        adata->items = g_slist_prepend (adata->items, adata->login_screen_item);
-        gtk_widget_show (adata->login_screen_item);
+                          G_CALLBACK (on_login_screen_activate), adata);
+        /* Only show switch user if there are other users */
 
-        adata->items = g_slist_sort_with_data (adata->items,
-                                               sort_menu_comparedatafunc,
-                                               adata);
+        adata->quit_session_item = gtk_image_menu_item_new_with_label (_("Quit..."));
+        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (adata->quit_session_item),
+                                       gtk_image_new ());
+        gtk_menu_shell_append (GTK_MENU_SHELL (adata->menu),
+                               adata->quit_session_item);
+        g_signal_connect (adata->quit_session_item, "style-set",
+                          G_CALLBACK (menuitem_style_set_cb), adata);
+        g_signal_connect (adata->quit_session_item, "activate",
+                          G_CALLBACK (on_quit_session_activate), adata);
+        gtk_widget_show (adata->quit_session_item);
 }
 
 static void
@@ -1022,7 +1091,7 @@ destroy_sub_menu (GdmAppletData *adata)
                                               G_CALLBACK (on_manager_user_added),
                                               adata);
         g_signal_handlers_disconnect_by_func (adata->manager,
-                                              G_CALLBACK (on_manager_user_is_logged_in_changed),
+                                              G_CALLBACK (on_manager_user_removed),
                                               adata);
 }
 
@@ -1073,9 +1142,13 @@ reset_icon (GdmAppletData *adata)
                 return;
         }
 
-        pixbuf = gdm_user_render_icon (adata->user, adata->panel_size);
-        if (pixbuf != NULL) {
-                image = gtk_image_menu_item_get_image (GTK_IMAGE_MENU_ITEM (adata->menuitem));
+        if (adata->user_item != NULL) {
+                image = gdm_entry_menu_item_get_image (GDM_ENTRY_MENU_ITEM (adata->user_item));
+                pixbuf = gdm_user_render_icon (adata->user, adata->panel_size * 3);
+                if (pixbuf == NULL) {
+                        return;
+                }
+
                 gtk_image_set_from_pixbuf (GTK_IMAGE (image), pixbuf);
                 g_object_unref (pixbuf);
         }
@@ -1093,6 +1166,7 @@ static void
 setup_current_user (GdmAppletData *adata)
 {
         const char *name;
+        GtkWidget  *label;
 
         adata->user = gdm_user_manager_get_user_by_uid (adata->manager, getuid ());
         if (adata->user != NULL) {
@@ -1103,11 +1177,12 @@ setup_current_user (GdmAppletData *adata)
         }
 
         adata->menuitem = gtk_image_menu_item_new_with_label (name);
-        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (adata->menuitem),
-                                       gtk_image_new ());
+        label = GTK_BIN (adata->menuitem)->child;
         gtk_menu_shell_append (GTK_MENU_SHELL (adata->menubar), adata->menuitem);
         gtk_widget_show (adata->menuitem);
 
+        update_label (adata);
+
         if (adata->user != NULL) {
                 reset_icon (adata);
 
@@ -1162,24 +1237,9 @@ fill_applet (PanelApplet *applet)
         adata->applet = applet;
         adata->panel_size = 24;
 
-        /* Until we add user selecting to GDM */
-        adata->active_only = TRUE;
-
         adata->client = gconf_client_get_default ();
 
-        tmp = g_strdup_printf ("applet-user-menu-item-%p", adata);
-        adata->user_menu_item_quark = g_quark_from_string (tmp);
-        g_free (tmp);
-
-        if (tooltips == NULL) {
-                tooltips = gtk_tooltips_new ();
-                g_object_ref (tooltips);
-                gtk_object_sink (GTK_OBJECT (tooltips));
-        } else {
-                g_object_ref (tooltips);
-        }
-
-        gtk_tooltips_set_tip (tooltips, GTK_WIDGET (applet), _("User Switcher"), NULL);
+        gtk_widget_set_tooltip_text (GTK_WIDGET (applet), _("Change account settings and status"));
         gtk_container_set_border_width (GTK_CONTAINER (applet), 0);
         gtk_widget_set_name (GTK_WIDGET (applet), "gdm-user-switch-applet");
         panel_applet_set_flags (applet, PANEL_APPLET_EXPAND_MINOR);
